/* Assumptions
This script assumes that each session is separated by 5 minutes (using Silverstein et al., 1999 statistics). A session can be a "single session" or a "multi interval session." A single session is one where a user submits 1 query and then does not not submit any other query within 5 minutes before or after that query. A multi interval session is one where a user submits successive queries within a 5 minute window. A single session has a total session time equal to 0 with a single query count. A multi interval session has a total session time equal to the time the first query was made in that session to the time the last query was made in that session. The total sessions, average of all sessions and maximum session lengths are all based on the above assumptions. So, there may be certain users who, for example, have only a single query. Those users will have a total session count of 1, and a total and average session time of 0.
*/

REGISTER ./tutorial.jar;
raw = LOAD 'excite-small.log' USING PigStorage('\t') AS (user, date:double, query);
clean1 = FILTER raw BY org.apache.pig.tutorial.NonURLDetector(query);
clean2 = FOREACH clean1 GENERATE user, date;

grouped1 = GROUP clean2 BY user; --group relation by user
grouped2 = FOREACH grouped1 GENERATE group as user, FLATTEN(clean2.date) as date1, clean2.date as dates2; --create 2 date columns which are copies. Flatten the first column to break up the grouping bag containing the dates

matrix1 = FOREACH grouped2 GENERATE user, date1, FLATTEN(dates2) as date2; --now break up the second grouping bag containing the same dates. Now have a matrix relation that matches each date to every other date for each user.
matrix2 = FILTER matrix1 BY date1 != date2; --remove the rows that have the same date

diff1 = FOREACH matrix2 GENERATE user, date1, date2, date2 - date1 as datediff; --take the difference between all the dates in the matrix
diff2 = FILTER diff1 BY datediff > 0; --remove the rows where the difference is < 0, which are duplicates of date pairings except in opposite order
diff3 = GROUP diff2 BY (user, date1); --now group by user and date1. This grouping makes sure you go back to the unique dates for each user
diff4 = FOREACH diff3 GENERATE FLATTEN(group), MIN(diff2.date2) as date2, MIN(diff2.datediff) as diff; --take the minimum of the date2 for the grouping bag, which would represent the next sequential date to date1. This will be called an "interval"
diff5 = FOREACH diff4 GENERATE group::user as user, group::date1 as date1, date2 as date2, diff as diff; --all intervals. This statement just cleans up diff4 for readability

diffgrouped = GROUP diff5 BY user; --group all intervals by users (will create a bag that lists all the intervals for each user)
diffclean1 = FOREACH diffgrouped GENERATE group, MIN(diff5.date1) as start, MAX(diff5.date2) as end, FLATTEN(diff5); --the "start date" and "end date" are the minimum and maximum of all dates in the interval, respecitively 
diffclean2 = FILTER diffclean1 BY diff5::diff > 500; --take only the intervals longer than 5 minutes
diffclean3 = FOREACH diffclean2 GENERATE group as user, start, end, diff5::date1 as mid1, diff5::date2 as mid2; --the intervals contained in this relation are those that have a difference between the dates that are longer than 1 hour. This represents a BREAK between sessions. The dates that mark the beginning and end of the BREAK are called "mid" dates (i.e. mid1 and mid2). Some users may not have this since they may have just had 1 session.

--separate out all the start dates, end dates and mid dates and put them in their own relation of {user, date}. Doing this so that it can be unioned together in the next step
diffclean_start = FOREACH diffclean1 GENERATE FLATTEN(group) as user, start as date;
diffclean_end = FOREACH diffclean1 GENERATE FLATTEN(group) as user, end as date;
diffclean_mid1 = FOREACH diffclean3 GENERATE user, mid1 as date;
diffclean_mid2 = FOREACH diffclean3 GENERATE user, mid2 as date;

--union all the start, end and mid dates. Then group them up in order to count the overlaps of the dates (meaning, the number of times a date is repeated per user). If a date is repeated in the union, that means that that specific date is a "singleSession" date, which means that the user queried once in that session and therefore does not have any interval of time between queries for that session.
union1 = union diffclean_start, diffclean_end, diffclean_mid1, diffclean_mid2;
union2 = GROUP union1 BY (user, date);

--lists user, date and number of times the date is repeated in sequence
overlap = FOREACH union2 GENERATE FLATTEN(group), COUNT(union1) as repeat;

--finds the total number of single sessions for each user
single_i_session1 = FILTER overlap BY repeat > 1;
single_i_session2 = GROUP single_i_session1 BY user;
single_i_session3 = FOREACH single_i_session2 GENERATE FLATTEN(group) AS user, COUNT(single_i_session1) as singleSessionCount; --this gives the total number of "singleSessions"
temp_allusers = FOREACH (GROUP clean2 BY user) GENERATE FLATTEN(group) as user, 1 as singleSessionCount; --all users have at least 1 session if they are logged. This will be overridden by the actual number if a person does NOT have exactly 1 non-empty query in the log
single_i_session4 = union temp_allusers, single_i_session3; --this is done to make sure that those users who didn't have their "singleSessions" counted are included in the relation. The users who didn't have a singleSession counted are those who queried only once. 
single_i_session_final = FOREACH (GROUP single_i_session4 BY user) GENERATE FLATTEN(group) AS user, MAX(single_i_session4.singleSessionCount) AS singleSessionCount; --take the maximum which represents to accurate number of "singleSessions" since there may be duplicates presented by the temp_allusers relation.

--find all the multi-sessions for each user (i.e. user, bag:multisessions((start, end, diff)))
multi_i_session1 = FILTER overlap BY repeat == 1; --the dates that did not repeat are considered dates that are part of a multi interval session, which means that the user made multiple queries in a given session.

--go through the process of making another date matrix of all the multi interval sessions in order to get a relation with the sequence of multi interval sessions
multi_i_session2 = FOREACH (GROUP multi_i_session1 BY user) GENERATE group AS user, FLATTEN(multi_i_session1.date) as date1, multi_i_session1.date as dates2;
multi_i_matrix1 = FOREACH multi_i_session2 GENERATE user, date1, FLATTEN(dates2) as date2;
multi_i_matrix2 = FILTER multi_i_matrix1 BY date1 != date2;
multi_i_matrix3 = FOREACH multi_i_matrix2 GENERATE user, date1, date2, date2 - date1 as diff;
multi_i_matrix4 = FILTER multi_i_matrix3 BY diff > 0;
multi_i_matrix5 = GROUP multi_i_matrix4 BY (user, date1);
multi_i_matrix6 = FOREACH multi_i_matrix5 GENERATE FLATTEN(group), MIN(multi_i_matrix4.date2) as date2, MIN(multi_i_matrix4.diff) as diff;
multi_i_matrix_final = FOREACH multi_i_matrix6 GENERATE group::user as user, group::date1 as date1, date2;

--The problem is that multi_i_matrix_final will have date1/date2 intervals that represent the BREAK between sessions. We want to remove these breaks from that matrix since they are not to be counted as a session. union multi_i_matrix_final with diffclean3 by the user so that you can find the date1/date2 pair that are NOT part of a session
dc3 = FOREACH diffclean3 GENERATE user, mid1 as date1, mid2 as date2;
multiunion1 = union multi_i_matrix_final, dc3;
multiunion2 = FOREACH (GROUP multiunion1 BY (user, date1, date2)) GENERATE group, COUNT(multiunion1) as repeat;--non-sessions should have a count that equals 2
multiunion3 = FILTER multiunion2 BY repeat == 1; --select only those intervals that are sessions. The dates that repeat == 2 are those that overlap with the BREAK mid dates;
multiunion4 = FOREACH multiunion3 GENERATE FLATTEN(group);
multiunion5 = FOREACH multiunion4 GENERATE group::user as user, group::date1 as date1, group::date2 as date2, group::date2 - group::date1 as diff;

multi_i_session_final = FOREACH (GROUP multiunion5 BY user) GENERATE FLATTEN(group) as user, multiunion5 as sessions;

--some users have no multi interval sessions since they just did single queries that are at least 1 hour apart from one another. Those users will have average session lengths equal to 0. The users with multi interval sessions will have their multisessions bag not == null. We have to separate these 2 types of users since the mathematics will not work for the singleSession users since they have a null multisessions bag.
sessions1 = join single_i_session_final by user LEFT, multi_i_session_final by user; --do a left join since the single_i_session_final relation will have all users included since it was joined with the temp_allusers relation. The multi_i_session_final relation will only have those users who have multi interval sessions
sessions2 = FOREACH sessions1 GENERATE single_i_session_final::user as user, single_i_session_final::singleSessionCount as singleSessionCount, multi_i_session_final::sessions as multisessions;

--for each of the 2 types of users (i.e. singleSession and multi interval session), find the totalSessions count, maxSessionLength and avgSessionLength (note - I'm not sure how to find the standard deviation since there is no square root function - i think)
sessions2_single = FOREACH (FILTER sessions2 BY multisessions is null) GENERATE user, singleSessionCount, multisessions, singleSessionCount as totalSessions, 0 as maxSessionLength, 0 as avgSessionLength;

sessions2_multi = FOREACH (FILTER sessions2 BY multisessions is not null) GENERATE user, singleSessionCount, multisessions, singleSessionCount + COUNT(multisessions) as totalSessions, MAX(multisessions.diff) as maxSessionLength, SUM(multisessions.diff)/(singleSessionCount + COUNT(multisessions)) as avgSessionLength;;

--Here is the final answers output
--describe answers > {user, totalSessions, maxSessionLength, avgSessionLength)
answers = FOREACH (UNION sessions2_single, sessions2_multi) GENERATE user, totalSessions, maxSessionLength, avgSessionLength;

dump answers;